from django.contrib.postgres import fields as psql_fields

from audoma.choices import make_choices
from audoma.django.db import models


COUNTRY_CHOICES = make_choices(
    "COUNTRIES",
    ((1, "PL", "Poland"), (2, "DE", "Germany")),
)


class ContactData(models.Model):

    phone_number = models.PhoneNumberField(region="PL")
    mobile = models.PhoneNumberField(region="PL")

    country = models.CharField(max_length=100, choices=COUNTRY_CHOICES.get_choices())
    city = models.CharField(max_length=200)


class Person(models.Model):
    class Meta:
        abstract = True

    name = models.CharField(max_length=255)
    surname = models.CharField(max_length=255)
    contact_data = models.ForeignKey(ContactData, on_delete=models.CASCADE)


class Specialization(models.Model):
    name = models.CharField(max_length=255)


class Doctor(Person):
    specialization = models.ManyToManyField(Specialization)
    salary = models.MoneyField(decimal_places=2, max_digits=10, default_currency="PLN")


class Patient(Person):
    weight = models.FloatField()
    height = models.DecimalField(decimal_places=2, max_digits=6)


class PatientFiles(models.Model):
    patient = models.ForeignKey(Patient, on_delete=models.CASCADE)

    @property
    def entries(self):
        return FileEntry.objects.filter(files=self)


class FileEntry(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    modifed_at = models.DateTimeField(auto_now=True)

    title = models.CharField(max_length=200)
    content = models.TextField()
    files = models.ForeignKey(PatientFiles, on_delete=models.CASCADE)


class Prescription(models.Model):

    issued_by = models.ForeignKey(Doctor, on_delete=models.CASCADE)
    issued_for = models.ForeignKey(Patient, on_delete=models.CASCADE)

    drugs = psql_fields.ArrayField(psql_fields.HStoreField())

    usable_in = psql_fields.DateRangeField()
    issued_in = psql_fields.CICharField(max_length=255)
    is_valid = models.BooleanField(default=True)



from healthcare_api import views

from audoma.drf import routers


# from django.contrib import admin
# from django.urls import (
#     include,
#     re_path,
# )


router = routers.DefaultRouter()

router.register(r"patient", views.PatientViewset, basename="patient")
router.register(
    r"specialization", views.SpecializatioNViewSet, basename="specialization"
)
router.register(r"doctor", views.DoctorViewset, basename="doctor")
router.register(r"prescritpion", views.PrescriptionViewSet, basename="prescription")


urlpatterns = [] + router.urls



# Generated by Django 4.1.3 on 2022-12-05 13:45

import django.contrib.postgres.fields
import django.contrib.postgres.fields.citext
import django.contrib.postgres.fields.hstore
import django.contrib.postgres.fields.ranges
import django.db.models.deletion
from django.contrib.postgres.operations import (
    CITextExtension,
    HStoreExtension,
)
from django.db import (
    migrations,
    models,
)

import audoma.django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ("healthcare_api", "0001_initial"),
    ]

    operations = [
        HStoreExtension(),
        CITextExtension(),
        migrations.CreateModel(
            name="Prescription",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "drugs",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=django.contrib.postgres.fields.hstore.HStoreField(),
                        size=None,
                    ),
                ),
                ("usable_in", django.contrib.postgres.fields.ranges.DateRangeField()),
                (
                    "issued_in",
                    django.contrib.postgres.fields.citext.CICharField(max_length=255),
                ),
                ("is_valid", audoma.django.db.fields.BooleanField(default=True)),
                (
                    "issued_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="healthcare_api.doctor",
                    ),
                ),
                (
                    "issued_for",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="healthcare_api.patient",
                    ),
                ),
            ],
        ),
    ]



from rest_framework.exceptions import ErrorDetail
from rest_framework.fields import CharField
from rest_framework.serializers import ValidationError
from rest_framework.test import APITestCase

from audoma.drf.fields import SerializerMethodField as AudomaSerializerMethodField


class SerializerMethodFieldTestCase(APITestCase):
    databases = "__all__"

    def test_create_without_field(self):
        field = AudomaSerializerMethodField()
        self.assertIsInstance(field, AudomaSerializerMethodField)
        self.assertIsNone(getattr(field, "field", None))

    def test_create_with_field(self):
        field = AudomaSerializerMethodField(
            field=CharField(max_length=255), writable=True
        )
        self.assertIsInstance(field, AudomaSerializerMethodField)
        self.assertIsNotNone(getattr(field, "field", None))
        self.assertIsInstance(field.field, CharField)

    def test_run_validation_valid_example(self):
        field = AudomaSerializerMethodField(
            field=CharField(max_length=255), writable=True
        )
        field.field_name = "test_field"
        self.assertEqual(field.run_validation("TestValue"), {"test_field": "TestValue"})
        self.assertEqual(field.run_validation(12), {"test_field": "12"})
        self.assertEqual(field.run_validation(21.37), {"test_field": "21.37"})

    def test_run_validation_invalid_example(self):
        field = AudomaSerializerMethodField(
            field=CharField(max_length=255), writable=True
        )
        try:
            field.run_validation("a" * 289)
        except ValidationError as e:
            self.assertEqual(
                e.detail[0],
                ErrorDetail(
                    "Ensure this field has no more than 255 characters.", "max_length"
                ),
            )

    def test_to_internal_value_success(self):
        field = AudomaSerializerMethodField(
            field=CharField(max_length=255), writable=True
        )
        field.field_name = "test_field"
        self.assertEqual(
            field.to_internal_value("TestValue"), {"test_field": "TestValue"}
        )
        self.assertEqual(field.to_internal_value(12), {"test_field": "12"})
        self.assertEqual(field.to_internal_value(21.37), {"test_field": "21.37"})

    def test_to_internal_value_fail_validation(self):
        field = AudomaSerializerMethodField(
            field=CharField(max_length=255), writable=True
        )
        try:
            field.to_internal_value("a" * 289)
        except ValidationError as e:
            self.assertEqual(
                e.detail[0],
                ErrorDetail(
                    "Ensure this field has no more than 255 characters.", "max_length"
                ),
            )

    def test_to_internal_value_fail_type(self):
        field = AudomaSerializerMethodField(
            field=CharField(max_length=255), writable=True
        )
        try:
            field.to_internal_value(object())
        except ValidationError as e:
            self.assertEqual(e.detail[0], ErrorDetail("Not a valid string.", "invalid"))

    def test_to_internal_value_raise_skipfield(self):
        field = AudomaSerializerMethodField(
            field=CharField(max_length=255), writable=True
        )
        try:
            field.to_internal_value(object())
        except ValidationError as e:
            self.assertEqual(e.detail[0], ErrorDetail("Not a valid string.", "invalid"))



from typing import (
    NamedTuple,
    Tuple,
    Union,
)

from django_filters import rest_framework as df_filters

from audoma.plumbing import create_choices_enum_description


class DocumentedTypedChoiceFilter(df_filters.TypedChoiceFilter):
    """Extended TypedChoiceFilter to generate documentation automatically"""

    def __init__(
        self, full_choices: Union[NamedTuple, Tuple], parameter_name: str, **kwargs
    ) -> None:
        self.parsed_choices = (
            full_choices.get_api_choices()
            if hasattr(full_choices, "get_choices")
            else full_choices
        )
        super().__init__(
            coerce=lambda value: full_choices.get_value_by_name(value),
            choices=self.parsed_choices,
            **kwargs,
        )
        self.full_choices = full_choices
        self.parameter_name = parameter_name
        self.extra["help_text"] = self.extra.get("help_text", "{choices}").format(
            choices=create_choices_enum_description(
                self.parsed_choices, self.field_name
            )
        )
        self.extra["choices"] = self.parsed_choices



from healthcare_api import models as api_models

from audoma.drf import serializers


class ContactDataSerializer(serializers.ModelSerializer):
    class Meta:
        model = api_models.ContactData
        fields = ["phone_number", "mobile", "country", "city"]


class PersonBaseSerializer(serializers.ModelSerializer):
    contact_data = ContactDataSerializer()

    def update(self, instance, validated_data):
        contact_data = validated_data.pop("contact_data", {})
        if not instance.contact_data:
            instance.contact_data = api_models.ContactData.objects.create(
                **contact_data
            )
        else:
            api_models.ContactData.objects.filter(pk=instance.contact_data.pk).update(
                **contact_data
            )
        instance = super().update(instance, validated_data)
        instance.contact_data.refresh_from_db()
        return instance

    def create(self, validated_data):
        contact_info = validated_data.pop("contact_data")
        validated_data["contact_data"] = api_models.ContactData.objects.create(
            **contact_info
        )
        return self.Meta.model.objects.create(**validated_data)


class PatientReadSerializer(PersonBaseSerializer):
    class Meta:
        model = api_models.Patient
        fields = ["name", "surname", "contact_data", "weight", "height"]


class PatientWriteSerializer(serializers.BulkSerializerMixin, PersonBaseSerializer):

    pk = serializers.IntegerField(write_only=True, required=False)

    class Meta:
        model = api_models.Patient
        fields = ["pk", "name", "surname", "contact_data", "weight", "height"]
        list_serializer_class = serializers.BulkListSerializer
        id_field = "pk"


class PatientFilesEntrySerializer(serializers.ModelSerializer):
    class Meta:
        model = api_models.FileEntry
        fields = ["created_at", "modifed_at", "title", "content"]


class PatientFilesDetailSerializer(serializers.ModelSerializer):

    patient = PatientReadSerializer()
    entries = PatientFilesEntrySerializer(many=True)

    class Meta:
        model = api_models.PatientFiles
        fields = ["patient", "entries"]


class SpecializationSerializer(serializers.ModelSerializer):
    class Meta:
        model = api_models.Specialization
        fields = ["id", "name"]


class DoctorReadSerializer(PersonBaseSerializer):

    specialization = SpecializationSerializer(many=True)

    class Meta:
        model = api_models.Doctor
        fields = ["name", "surname", "contact_data", "specialization"]


class DoctorWriteSerializer(PersonBaseSerializer):

    choices_options_links = {
        "specialization": {
            "viewname": "specialization-list",
            "value_field": "id",
            "display_field": "name",
        }
    }

    specialization = serializers.SerializerMethodField(
        field=serializers.ListField(child=serializers.IntegerField()), writable=True
    )

    def get_specialization(self, doctor):
        return doctor.specialization.values_list("id", flat=True)

    def validate(self, attrs):
        self.specialization_pks = attrs.pop("specialization", [])
        return attrs

    def save(self, **kwargs):
        instance = super().save(**kwargs)
        for s in api_models.Specialization.objects.filter(
            pk__in=self.specialization_pks
        ):
            instance.specialization.add(s)
        instance.save()
        return instance

    class Meta:
        model = api_models.Doctor
        fields = ["name", "surname", "contact_data", "specialization", "salary"]


class PerscriptionReadSerializer(serializers.ModelSerializer):
    class Meta:
        model = api_models.Prescription
        fields = [
            "issued_by",
            "issued_for",
            "drugs",
            "usable_in",
            "issued_in",
            "is_valid",
        ]


class PrescrtiptionWriteSerializer(serializers.ModelSerializer):
    class Meta:
        model = api_models.Prescription
        fields = ["issued_by", "issued_for", "drugs", "usable_in", "issued_in"]



from drf_spectacular.views import (
    SpectacularAPIView,
    SpectacularRedocView,
    SpectacularSwaggerView,
)

from django.contrib import admin
from django.urls import (
    include,
    re_path,
)


urlpatterns = (
    re_path("v1/", include("drf_example.v1_urls")),
    re_path("v2/", include("drf_example.v2_urls")),
    re_path("admin/", admin.site.urls),
    re_path(r"^schema/$", SpectacularAPIView.as_view(), name="schema"),
    re_path(
        r"^swagger-ui/$",
        SpectacularSwaggerView.as_view(url_name="schema"),
        name="swagger-ui",
    ),
    re_path(r"^redoc/$", SpectacularRedocView.as_view(url_name="schema"), name="redoc"),
)
